
# torch imports
import lightning as l
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import CSVLogger

class NicheViT(l.LightningModule):

    def __init__(
        self,
        loss,  # options: [MSE, MAE]
        optimizer,  # options: [Adam, SGD]
        lr,  # learning rate
    ):
        pass


"""
A class to organize YOLO-structured data

Methods
---

get_images
    list of absolute paths of images in root/<split>/images
clone
    copy root/train and root/test to root/<folder_name>
shuffle_train_val
    shuffle self.ls_train_images and assign to
save_yaml
save_txt


Folder structure
---
root/
    train/ (required)
        images/
            img_1_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.jpg
            img_2_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.jpg
            ...
        labels/
            img_1_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.txt
            img_2_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.txt
            ...
    test/ (required)
        images/
            img_3.jpg
            img_4.jpg
            ...
        labels/
            img_3.txt
            img_4.txt

    test.txt (generated)
    train.txt (generated)
    val.txt (generated)
    data.yaml (generated)

Example YAML
---
path: /home/niche/cowsformer/data/cow200/yolov5/run3
train: "train.txt"
val: "val.txt"
test: "test.txt"
names:
  0: none
  1: cow

Example train.txt
---

/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_32_jpg
/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_1_jpg
/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_1_26_jpg
/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_1_62_jpg
/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_1_10_jpg
/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_3_11_jpg
/home/niche/cowsformer/data/cow200/yolov5/run0/images/img_4_3_jpg
"""

import os
import shutil
import random
import PIL
import numpy as np
import supervision as sv

# local imports
from pyniche.data.bbox import xywh2xyxy


class YOLO_API:
    def __init__(
        self,
        root: str,
    ):
        self.root = root
        self.ls_train_images_all = self.get_images("train")
        self.ls_train_images = None
        self.ls_val_iamges = None
        self.ls_test_images = self.get_images("test")
        self.save_txt("test")

    def get_PIL(self, split, idx):
        """
        get PIL image from split and idx

        params
        ------
        split: str
            "train" or "test"
        idx: int
            index of the image

        return
        ------
        PIL.Image
        """
        if split == "train":
            path = self.ls_train_images_all[idx]
        else:
            path = self.ls_test_images[idx]
        return PIL.Image.open(path)

    def get_images(self, split):
        """
        search images names (.jpg) in root/<split>/images

        params
        ------
        split: str
            "train" or "test"

        return
        ------
        a list of aboslute paths of images
        """
        return self.get_filepaths(split, "images")

    def get_labels(self, split):
        """
        search label names (.txt) in root/<split>/labels

        params
        ------
        split: str
            "train" or "test"

        return
        ------
        a list of aboslute paths of labels
        """
        return self.get_filepaths(split, "labels")

    def get_filepaths(self, split, folder):
        """
        get file paths of images or labels

        params
        ---
        split: str
            "train" or "test"
        folder: str
            "images" or "labels"
        """
        path_files = os.path.join(self.root, split, folder)
        ext = ".jpg" if folder == "images" else ".txt"
        ls_files = [f for f in os.listdir(path_files) if f.endswith(ext)]
        ls_files = [os.path.join(path_files, f) for f in ls_files]
        return sorted(ls_files)

    def get_detections(self, split):
        """
        get sv.Detections from labels

        params
        ------
        split: str
            "train" or "test"

        return
        ------
        a list of sv.Detections
        """
        detections = []
        # get file paths
        labels = self.get_labels(split)
        images = self.get_images(split)
        n_samples = len(images)
        # iterate each pair of image and label
        for i in range(n_samples):
            # get image info
            image = PIL.Image.open(images[i])
            img_w, img_h = image.size
            # get annotation
            label = labels[i]
            with open(label, "r") as f:
                lines = f.readlines()
                lines = [l.strip() for l in lines]
            # each detection in the image/label
            ls_xyxy = []
            ls_cls = []
            for l in lines:
                # each line: [class_id, x_center, y_center, width, height]
                class_id, x_center, y_center, width, height = l.split(" ")
                class_id = int(class_id)
                xyxy = xywh2xyxy(
                    (x_center, y_center, width, height),
                    img_size=(img_w, img_h),
                )
                ls_xyxy.append(xyxy)
                ls_cls.append(class_id)
            ls_xyxy = np.array(ls_xyxy)
            ls_cls = np.array(ls_cls)
            detections.append(sv.Detections(ls_xyxy, class_id=ls_cls))
        return detections

    def clone(self, folder_name):
        """
        copy root/train and root/test to
        root/<folder_name>/train and root/<folder_name>/test
        """
        path_train = os.path.join(self.root, "train")
        path_test = os.path.join(self.root, "test")
        path_folder = os.path.join(self.root, folder_name)
        if os.path.exists(path_folder):
            shutil.rmtree(path_folder)
        os.mkdir(path_folder)
        shutil.copytree(path_train, os.path.join(path_folder, "train"))
        shutil.copytree(path_test, os.path.join(path_folder, "test"))
        # copy yaml and other txt
        shutil.copy(os.path.join(self.root, "data.yaml"), path_folder)
        shutil.copy(os.path.join(self.root, "train.txt"), path_folder)
        shutil.copy(os.path.join(self.root, "test.txt"), path_folder)

    def shuffle_train_val(self, n=None, k=5):
        """
        shuffle self.ls_train_images and assign to
        self.ls_train_images and self.ls_val_images

        params
        ------
        n: None or int or float
            None: use all images
            int: number of images to be included in the train/val set
            float: ratio of images to be included in the train/val set
        k: int
            how many folds to split the train/val set
        """
        # determine n
        total_n = len(self.ls_train_images_all)
        if n is None:
            n = total_n
        elif isinstance(n, float):
            n = int(n * total_n)
        n_val = int(n / k)
        # shuffle training images
        random.shuffle(self.ls_train_images_all)
        train_images = self.ls_train_images_all[:n]
        # split train/val
        self.ls_train_images = train_images[:-n_val]

        self.ls_val_images = train_images[-n_val:]
        self.save_txt("train")
        self.save_txt("val")

    def save_yaml(self, classes, name="data.yaml"):
        """
        make data.yaml in root

        params
        ------
        classes: list
            e.g., ["cow", "none"]

        name: str
            name of the yaml file

        """
        path_yaml = os.path.join(self.root, name)
        with open(path_yaml, "w") as f:
            f.write(f"path: {self.root}\n")
            f.write(f'train: "train.txt"\n')
            f.write(f'val: "val.txt"\n')
            f.write(f'test: "test.txt"\n')
            f.write("names:\n")
            for i, c in enumerate(classes):
                f.write(f"  {i}: {c}\n")

    def save_txt(self, split):
        """
        save <split>.txt in root
        """
        path_txt = os.path.join(self.root, f"{split}.txt")
        with open(path_txt, "w") as f:
            for img in getattr(self, f"ls_{split}_images"):
                f.write(img + "\n")



import os
import json
from PIL import Image
import datasets


def hf_to_yolo(
    datadict,
    dir_dest,
    classes=["category 1"],
    size_new=None,
):
    """
    iteratively extract Dataset from DatasetDict and save to YOLO format

    param
    ----
    datadict: DatasetDict
    dir_dest: str
        destination folder
    size_new: tuple
        (width, height)
    classes: list
        list of class names

    output
    ---
    dir_dest/
        data.yaml
        <split_1>.txt
        <split_2>.txt
        <split_1>/
            images/
                image_1.jpg
                image_2.jpg
                ...
            labels/
                image_1.txt
                image_2.txt
                ...
        <split_2>/
            ...
    """
    splits = []
    for split in datadict:
        print(f"Processing split: {split}")
        splits.append(split)
        split_dir = os.path.join(dir_dest, split)
        split_file = os.path.join(dir_dest, f"{split}.txt")
        filenames = _hf_to_yolo(datadict[split], split_dir, size_new)
        with open(split_file, "a") as f:
            for filename in filenames:
                f.write(os.path.join(split_dir, "images", filename) + "\n")

    # Create data.yaml file
    yaml_path = os.path.join(dir_dest, "data.yaml")
    with open(yaml_path, "w") as yaml_file:
        yaml_file.write("path: %s\n" % dir_dest)
        yaml_file.write("nc: %d\n" % len(classes))
        # splits
        for split in splits:
            yaml_file.write(f"{split}: {split}.txt\n")
        # classes
        yaml_file.write("names:\n")
        for i, class_name in enumerate(classes):
            yaml_file.write(f"  {i}: {class_name}\n")


def _hf_to_yolo(
    dataset,
    dir_dest,
    size_new=None,
):
    """
    turn one dataset instance to YOLO format and save to dir_dest

    param
    ----
    dataset: Dataset instance
    dir_dest: str
        destination folder
    size_new: tuple
        (width, height)

    output
    ---
    dir_dest/
        images/
            image_1.jpg
            image_2.jpg
            ...
        labels/
            image_1.txt
            image_2.txt
    """
    # Create directories
    os.makedirs(os.path.join(dir_dest, "images"), exist_ok=True)
    os.makedirs(os.path.join(dir_dest, "labels"), exist_ok=True)

    # extract dataset
    ls_annotations = dataset["annotations"]
    ls_filename = dataset["filename"]
    ls_image = dataset["image"]

    # Process each image and annotation
    for i in range(len(dataset)):
        # log
        filename = ls_filename[i]
        print("(%d/%d) %s\r" % (i + 1, len(dataset), filename))

        # image
        image_path = os.path.join(dir_dest, "images", filename)
        image = ls_image[i]
        size_ori = image.size
        if size_new:
            image = image.resize(size_new, Image.Resampling.LANCZOS)
        else:
            size_new = size_ori
        image.save(image_path)

        # label
        label_path = os.path.join(dir_dest, "labels", filename.replace(".jpg", ".txt"))
        with open(label_path, "w") as label_file:
            for annotation, bbox in zip(
                ls_annotations[i]["category_id"],
                ls_annotations[i]["bbox"],
            ):
                # Scale bbox coordinates
                scale_x = size_new[0] / size_ori[0]
                scale_y = size_new[1] / size_ori[1]
                scaled_bbox = [
                    bbox[0] * scale_x,
                    bbox[1] * scale_y,
                    bbox[2] * scale_x,
                    bbox[3] * scale_y,
                ]

                # Convert scaled bbox to YOLO format
                x_center = (scaled_bbox[0] + scaled_bbox[2] / 2) / size_new[0]
                y_center = (scaled_bbox[1] + scaled_bbox[3] / 2) / size_new[1]
                width = scaled_bbox[2] / size_new[0]
                height = scaled_bbox[3] / size_new[1]

                # Write to label file
                # minus 1 because YOLO uses 0-based class index
                label_file.write(
                    f"{annotation - 1} {x_center} {y_center} {width} {height}\n"
                )
    # return filenames
    return ls_filename


def hf_to_coco(
    datadict,
    dir_dest,
    classes=["category 1"],
    size_new=None,
):
    """
    param
    ---
    datadict: DatasetDict
    dir_dest: str
        destination folder
    size_new: tuple
        (width, height)
    classes: list
        list of class names

    output
    ---
    dir_dest/
        <split_1>/
            data.json
            image_1.jpg
            image_2.jpg
            ...
        <split_2>/
            ...
        ...
    """
    splits = []
    for split in datadict:
        print(f"Processing split: {split}")
        splits.append(split)
        split_dir = os.path.join(dir_dest, split)
        os.makedirs(split_dir, exist_ok=True)
        _hf_to_coco(datadict[split], split_dir, classes, size_new)


def _hf_to_coco(
    dataset,
    dir_dest,
    classes=["category 1"],
    size_new=None,
):
    """
    param
    ---
    dataset: Dataset instance
    dir_dest: str
        destination folder
    size_new: tuple
        (width, height)
    classes: list
        list of class names

    output
    ---
    dir_dest/
        data.json
        image_1.jpg
        image_2.jpg
        ...
    """
    coco_format = {
        "info": {},
        "licenses": [],
        "categories": [],
        "images": [],
        "annotations": [],
    }
    # extract dataset
    ls_annotations = dataset["annotations"]
    ls_imgage_id = dataset["image_id"]
    ls_filename = dataset["filename"]
    ls_image = dataset["image"]

    category_set = set()
    for i in range(len(dataset)):
        # Process categories
        for category_id in ls_annotations[i]["category_id"]:
            category_set.add(category_id)

    # Add categories
    for i, category_id in enumerate(category_set):
        try:
            coco_format["categories"].append(
                {
                    "id": category_id,
                    "name": classes[i],
                    "supercategory": classes[i],
                }
            )
        except:
            raise ValueError(
                f"Number of classes ({len(classes)}) should match the number of unique categories ({len(category_set)})"
            )

    annotation_id = 1  # Initialize annotation ID
    for i in range(len(dataset)):
        image_id = ls_imgage_id[i]
        filename = ls_filename[i]
        image = ls_image[i]
        img_w, img_h = image.size

        print("(%d/%d) %s\r" % (i + 1, len(dataset), filename))

        # Resize image if size_new is provided
        if size_new:
            image = image.resize(size_new, Image.Resampling.LANCZOS)
        image.save(os.path.join(dir_dest, filename))

        # Add image information
        coco_format["images"].append(
            {
                "id": image_id,
                "width": image.width,
                "height": image.height,
                "file_name": filename,
            }
        )

        # Add annotations
        for category_id, bbox in zip(
            ls_annotations[i]["category_id"], ls_annotations[i]["bbox"]
        ):
            # Scale bbox if image is resized
            if size_new:
                scale_x = size_new[0] / img_w
                scale_y = size_new[1] / img_h
                scaled_bbox = [
                    bbox[0] * scale_x,  # x
                    bbox[1] * scale_y,  # y
                    bbox[2] * scale_x,  # w
                    bbox[3] * scale_y,  # h
                ]
            else:
                scaled_bbox = bbox

            coco_format["annotations"].append(
                {
                    "id": annotation_id,
                    "image_id": image_id,
                    "category_id": category_id,
                    "bbox": scaled_bbox,
                    "area": scaled_bbox[2] * scaled_bbox[3],  # width * height
                    "iscrowd": 0,  # Assuming individual objects
                }
            )
            annotation_id += 1
    # Write to JSON file
    with open(os.path.join(dir_dest, "data.json"), "w") as f:
        json.dump(coco_format, f, indent=4)



import sys
import os
import json
import random
from PIL import Image

# torch imports
import torchvision

os.chdir("..")
sys.path.append(".")

ROOT = os.getcwd()


class CocoDetectionDataset(torchvision.datasets.CocoDetection):
    def __init__(self, root, file_json, processor):
        """
        prerequisite: json file must be placed in root/labels/
        args:
            root: root directory of the dataset
            file_json: json file name, e.g. _coco_train.json
            processor: pre-trained transformation pipeline (can it be other than DETR?)
        """
        self.root = root
        self.processor = processor
        path_json = os.path.join(root, "labels", file_json)
        super().__init__(root, path_json)

    def _load_image(self, id: int) -> Image.Image:
        # optional
        image_name = self.coco.loadImgs(id)[0]["file_name"]
        image_path = os.path.join(self.root, "images", image_name)
        image = Image.open(image_path).convert("RGB")
        return image

    def __getitem__(self, idx):
        # read in PIL image and target in COCO format
        # feel free to add data augmentation here before passing them to the next step
        img, target = super().__getitem__(idx)

        return img, target
        # # preprocess image and target (converting target to DETR format, resizing + normalization of both image and target)
        # image_id = self.ids[idx]
        # target = {"image_id": image_id, "annotations": target}
        # encoding = self.processor(images=img, annotations=target, return_tensors="pt")
        # pixel_values = encoding["pixel_values"].squeeze()  # remove batch dimension
        # target = encoding["labels"][0]  # remove batch dimension

        # return pixel_values, target


from transformers import DetrForObjectDetection, DetrImageProcessor

root_data = os.path.join(
    ROOT,
    "pyniche",
    "data",
    "examples",
    "balloon",
)
path_model = "facebook/detr-resnet-50"
processor = DetrImageProcessor.from_pretrained(path_model)
dataset = CocoDetectionDataset(root_data, "_coco_test.json", processor=processor)
idx = 1
img, target = dataset[1]
coco = dataset.coco



# get filename
ids = coco.getImgIds()
meta_img = coco.loadImgs(ids)

meta = meta_img[0]
filename = meta["file_name"]
img_id = meta["id"]
ann_ids = coco.getAnnIds(img_id)
annotations = coco.loadAnns(ann_ids)

annotation = annotations[0]
annotation["id"]
annotation["category_id"]
coco.cats[annotation["category_id"]]["supercategory"]
annotation["bbox"]
annotation["area"]


len(annotations)
meta
coco.getImgIds()


coco.imgs.keys()

# len
coco.loadImgs(1)
coco.showAnns(coco.loadAnns(1))
coco.loadAnns(1)


image_id = dataset.ids[idx]

target = {"image_id": image_id, "annotations": target}
encoding = processor(images=img, annotations=target, return_tensors="pt")
target

encoding.keys()
len(encoding["labels"])
pixel_values.size()
encoding["pixel_values"].size()
pixel_values = encoding["pixel_values"].squeeze()  # remove batch dimension
target = encoding["labels"][0]  # remove batch dimension
target

features=datasets.Features(
    {
        "image": datasets.Image(),
        "annotations": datasets.Sequence(
            {
                "id": datasets.Value("int64"),
                "supercategory": datasets.ClassLabel(names=COCO_SUPCLASSES),
                "category": datasets.ClassLabel(names=COCO_CLASSES),
                "bbox": datasets.Sequence(datasets.Value("float64"), length=4),
                "area": datasets.Value("float64"),
            }),
        "image_id": datasets.Value("int64"),
        "filename": datasets.Value("string"),
    }
),
homepage="githu


import torch


def get_device():
    """Get device (cuda, mps, or cpu)"""
    if torch.backends.mps.is_available():
        device = "mps"
    elif torch.cuda.is_available():
        device = "0"
    else:
        device = "cpu"
    return device



"""
A class to modify COCO json files.

methods
---
- i/o
    - load
    - save

- subsetting
    - subset_by_dir
    - subset_by_image_ids

- getters
    - get_filenames_from_dir
    - get_ids_from_imgs
    - get_img_by_filename
    - get_img_by_image_id
    - get_ann_by_image_id

- concatenate
"""

# native imports
import json
import os
import numpy as np
import PIL

# local imports
from pyniche.data.bbox import xywh2xyxy

# deep learning
import supervision as sv


class COCO_API:

    def __init__(
        self,
        path_json=None,
        data=None,
    ):
        # members
        self.data = data  # user-provided dict of COCO keys
        self.path_json = path_json
        # init
        if not data:
            # 2. user provided path to json
            self.load()

        # load filepaths
        self.filepaths = self.get_filepaths_from_imgs(self.images())
        self.image_ids = self.get_ids_from_imgs(self.images())

    def load(self):
        with open(self.path_json, "r") as f:
            self.data = json.load(f)

    def save(self, path):
        with open(path, "w") as f:
            json.dump(self.data, f)

    def new_instance(self, imgs_coco, ann_coco):
        # re-index images and annotations
        # imgs_coco, ann_coco = reindex_coco(imgs_coco, ann_coco)

        # new instance
        new_json = dict(
            {
                "info": self.info(),
                "licenses": self.licenses(),
                "categories": self.categories(),
                "images": imgs_coco,
                "annotations": ann_coco,
            }
        )
        return COCO_API(data=new_json)

    # subsetting
    def subset_by_dir(self, dir_img):
        # get images from dir
        filenames = self.get_filenames_from_dir(dir_img)
        imgs_coco = self.get_img_by_filename(filenames)

        # get ann by ids of images
        ids = self.get_ids_from_imgs(imgs_coco)
        ann_coco = self.get_ann_by_image_id(ids)

        # return
        return self.new_instance(imgs_coco, ann_coco)

    def subset_by_image_ids(self, ids):
        if isinstance(ids, int):
            # if int, it's a single id
            ids = [ids]
        elif isinstance(ids, list):
            # if list, it's a list of ids
            ids = ids
        elif isinstance(ids, tuple):
            # if tuple, it's a range
            ids = list(np.arange(ids[0], ids[1] + 1))

        # filter images
        imgs_coco = self.get_img_by_image_id(ids)
        ann_coco = self.get_ann_by_image_id(ids)

        # return
        return self.new_instance(imgs_coco, ann_coco)

    # getters
    def get_detections(self):
        """
        get sv.Detections from COCO annotations

        return
        ------
        a list of sv.Detections
        """
        detections = []
        img_ids = self.get_ids_from_imgs(self.images())
        for id in img_ids:
            anns = self.get_ann_by_image_id([id])
            ls_xyxy = []
            ls_cls = []
            for ann in anns:
                xywh = ann["bbox"]
                cls = ann["category_id"]
                xyxy = xywh2xyxy(xywh, in_xy="top-left")
                ls_xyxy.append(xyxy)
                ls_cls.append(cls)
            ls_xyxy = np.array(ls_xyxy)
            ls_cls = np.array(ls_cls)
            detections.append(sv.Detections(ls_xyxy, class_id=ls_cls))
        return detections

    def get_PIL(self, idx):
        filepath = self.filepaths[idx]
        img = PIL.Image.open(filepath)
        return img

    def get_filepaths_from_imgs(self, imgs_coco):
        path_dir = os.path.dirname(self.path_json)
        filepaths = [os.path.join(path_dir, i["file_name"]) for i in imgs_coco]
        return filepaths

    def get_filenames_from_dir(self, dir_img):
        filenames = [f for f in os.listdir(dir_img) if ".jpg" in f or ".png" in f]
        return sorted(filenames)

    def get_ids_from_imgs(self, imgs_coco):
        ids = [i["id"] for i in imgs_coco]
        return ids

    def get_img_by_filename(self, filenames):
        imgs_coco = [i for i in self.images() if i["file_name"] in filenames]
        return imgs_coco

    def get_img_by_image_id(self, ids):
        imgs_coco = [i for i in self.images() if i["id"] in ids]
        return imgs_coco

    def get_ann_by_image_id(self, ids):
        ann_coco = [a for a in self.annotations() if a["image_id"] in ids]
        return ann_coco

    # concat
    def concatenate(self, coco):
        """
        both self and coco are COCO_API instances
        """
        # check the self data and max id
        imgs_self = self.images()
        ann_self = self.annotations()

        # check the coco data and reinex
        imgs_coco = coco.images()
        ann_coco = coco.annotations()
        # add 10000 to the coco id
        for i, img in enumerate(imgs_coco):
            imgs_coco[i]["id"] += 10000
        for i, ann in enumerate(ann_coco):
            ann_coco[i]["id"] += 10000
            ann_coco[i]["image_id"] += 10000
        # concatenate
        imgs_new = imgs_self + imgs_coco
        ann_new = ann_self + ann_coco
        return self.new_instance(imgs_new, ann_new)

    # COCO keys
    def info(self):
        return self.data["info"]

    def licenses(self):
        return self.data["licenses"]

    def categories(self):
        return self.data["categories"]

    def images(self):
        return self.data["images"]

    def annotations(self):
        return self.data["annotations"]



# native imports
import os

# torch imports
import torch
import lightning as L
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import TensorBoardLogger


class NicheTrainer:
    def __init__(self, device="cpu"):
        # core
        self.model = None  # lightning module
        self.data = None  # lightning data module
        self.trainer = None  # lightining trainer
        self.device = device  # cpu, cuda, or mps
        # outputs
        self.loggers = None  # lightning loggers
        self.callbacks = None  # lightning callbacks
        self.out = dict(
            {
                "dir": None,
                "best_loss": None,
                "best_path": None,
            }
        )

    def set_model(
        self,
        model_class: L.LightningModule,
        checkpoint=None,
        **kwargs,
    ):
        """
        parameters
        ---
        model_class: l.LightningModule
            the lightning module class, e.g., transformers.DetrModel

        other keyword arguments
        ---
        pretrained: str
            path to the pretrained model, e.g., facebook/detr-resnet-50
        checkpoint: str
            local path to the checkpoint, e.g., model.ckpt
        config: any
            model configuration, e.g., transformers.DetrConfig

        """
        if checkpoint:
            self.model = model_class.load_from_checkpoint(checkpoint, **kwargs)
            print(f"model loaded from {checkpoint}")
        else:
            # pretrained or config
            self.model = model_class(**kwargs)
        self.model.to(self.device)

    def set_data(
        self,
        dataclass: L.LightningDataModule,
        **kwargs,  # varied arguments for the dataclass
    ):
        self.data = dataclass(**kwargs)

    def set_out(
        self,
        dir_out: str,
    ):
        self.out["dir"] = dir_out
        if not os.path.exists(self.out["dir"]):
            os.makedirs(self.out["dir"])

    def fit(
        self,
        epochs: int = 100,
    ):
        self.loggers = get_logger(self.out["dir"])
        self.callbacks = get_checkpoint(self.out["dir"])
        self.trainer = L.Trainer(
            max_epochs=epochs,
            callbacks=self.callbacks,
            logger=self.loggers,
        )
        self.trainer.fit(self.model, self.data)

    def val(self):
        self.trainer = L.Trainer()
        out = self.trainer.validate(self.model, self.data)
        return out

    def predict(self, split="test"):
        """
        this will call model.predict_step()
        """
        self.trainer = L.Trainer()
        if split == "train":
            dataloader = self.data.train_dataloader()
        elif split == "val":
            dataloader = self.data.val_dataloader()
        elif split == "test":
            dataloader = self.data.test_dataloader()
        out = self.trainer.predict(self.model, dataloader)
        # out is a nested list (# of batches, batch_size)
        # flatten the list
        out = [item for sublist in out for item in sublist]
        return out

    def get_best_loss(
        self,
        rm_threshold: float = 1e5,
    ):
        self.out["best_loss"] = self.callbacks.best_model_score.item()
        self.out["best_path"] = self.callbacks.best_model_path
        if self.out["best_loss"] > rm_threshold:
            os.remove(self.out["best_path"])
        return self.out["best_loss"]

    def load_best_model(self):
        try:
            best_path = self.callbacks.best_model_score.item()
            self.model = self.model.load_from_checkpoint(best_path)
            self.model.to(self.device)
            print(f"model loaded from {best_path}")
        except Exception as e:
            print(e)


def get_logger(dir_out):
    # training configuration
    logger = TensorBoardLogger(
        save_dir=dir_out,
        name=".",  # will not create a new folder
        version=".",  # will not create a new folder
        log_graph=True,  # for model architecture visualization
        default_hp_metric=False,
    )  # output: save_dir/name/version/hparams.yaml
    return logger


def get_checkpoint(dir_out):
    checkpoint_callback = ModelCheckpoint(
        monitor="val_loss",
        dirpath=dir_out,
        mode="min",
        save_top_k=1,
        verbose=False,
        save_last=False,
        filename="model-{val_loss:.3f}",
    )
    return checkpoint_callback



import lightning as L

from torch.utils.data import random_split, DataLoader
from datasets import concatenate_datasets, load_dataset


class BaseDataModule(L.LightningDataModule):
    """
    parameters
    ---
    dataname and configname: str
        local config file or repository name on the huggingface hub
    batch: int (default: 32, optional)
        batch size
    n_train: int or float (default: None, optional)
        if specified, use only n_train samples in the train/val process
        otherwise, use the default split
    """

    def __init__(
        self,
        dataname: str,
        configname: str = None,
        batch: int = 32,
        n_train: any = None,  # no redistribution if None
        k: int = 5,  # k-fold cross validation
        **kwargs,
    ):
        super().__init__()
        # input parameters
        self.dataname = dataname
        self.configname = configname
        self.batch = batch
        self.n_train = n_train
        self.k = k
        # datasets
        self.dataset = {
            "train": None,
            "val": None,
            "test": None,
        }
        self.kwargs = kwargs

    def prepare_data(self):
        # download
        pass

    def setup(self, stage: str):
        # Assign train/val datasets for use in dataloaders
        if stage == "fit":
            # load and re-distribute train/val
            data_train = self.get_dataset("train")
            try:
                data_val = self.get_dataset("validation")
                data_train, data_val = self.re_distribute(
                    data_train,
                    data_val,
                    self.n_train,
                    self.k,
                )
            except:
                # if the dataset does not have a validation split
                data_train, data_val = self.re_distribute(
                    data_train,
                    None,
                    self.n_train,
                    self.k,
                )
            # assign and set transforms
            self.dataset["train"] = data_train
            self.dataset["val"] = data_val
        elif stage == "test":
            data_test = self.get_dataset("test")
            self.dataset["test"] = data_test

    def get_dataset(self, split):
        return load_dataset(
            self.dataname,
            self.configname,
            split=split,
            **self.kwargs,
        )

    def set_dataset(self, split):
        """
        split: str
            train, val, or test
        """
        self.dataset[split] = self.get_dataset(split)

    # loaders
    def get_dataloader(self, split):
        if split == "train":
            return self.train_dataloader()
        elif split == "val":
            return self.val_dataloader()
        elif split == "test":
            return self.test_dataloader()

    def train_dataloader(self):
        if self.dataset["train"] is None:
            self.set_dataset("train")
        return DataLoader(
            self.dataset["train"].with_transform(self._transform_train),
            batch_size=self.batch,
            shuffle=True,
            pin_memory=True,
            collate_fn=self._collate_fn_train,
        )

    def val_dataloader(self):
        if self.dataset["val"] is None:
            self.set_dataset("val")
        return DataLoader(
            self.dataset["val"].with_transform(self._transform_val),
            batch_size=self.batch,
            shuffle=False,
            pin_memory=True,
            collate_fn=self._collate_fn_val,
        )

    def test_dataloader(self):
        if self.dataset["test"] is None:
            self.set_dataset("test")
        return DataLoader(
            self.dataset["test"].with_transform(self._transform_test),
            batch_size=self.batch,
            shuffle=False,
            pin_memory=True,
            collate_fn=self._collate_fn_test,
        )

    # additional methods
    def re_distribute(self, data_train, data_val=None, n=None, k=5):
        """
        data_train, data_val: datasets.Dataset
        n: int or float, sample size in train and val
            if None, do nothing
            if -1, use all samples
            if int, number of samples in train and val
            if float, ratio of samples in train and val
        k: int
            how many folds to split the train/val set

        """
        if n is not None:
            # concatenate to get the full dataset
            if data_val is None:
                data_full = concatenate_datasets([data_train, data_val])
            else:
                data_full = data_train

            # determine n
            total_n = len(data_full)
            if n == -1:
                n = total_n
            elif isinstance(n, float):
                n = int(n * total_n)
            n_val = int(n / k)

            # split
            data_train, data_val = random_split(data_full, [n - n_val, n_val])

        # return
        return data_train, data_val

    # need to be implemented
    def _collate_fn(self, batch):
        raise NotImplementedError

    def _collate_fn_train(self, batch):
        batch = self._collate_fn(batch)
        return batch

    def _collate_fn_val(self, batch):
        batch = self._collate_fn(batch)
        return batch

    def _collate_fn_test(self, batch):
        batch = self._collate_fn(batch)
        return batch

    def _transform_train(examples):
        return examples

    def _transform_val(examples):
        return examples

    def _transform_test(examples):
        return examples



"""
# Reference:
https://lightning.ai/docs/pytorch/stable/common/lightning_module.html
https://github.com/NielsRogge/Transformers-Tutorials/blob/master/DETR/Fine_tuning_DetrForObjectDetection_on_custom_dataset_(balloon).ipynb
"""

# torch imports
import torch
from transformers import DetrForObjectDetection, DetrImageProcessor, DetrConfig
import lightning as L
import supervision as sv

PRESET_MODEL = "facebook/detr-resnet-50"


class Niche_Detr(L.LightningModule):
    def __init__(
        self,
        n_classes=1,
        lr=1e-4,
        lr_backbone=1e-5,
        weight_decay=1e-4,
        # model weights
        pretrained=PRESET_MODEL,  # facebook/detr-resnet-50, facebook/detr-resnet-101
        # others
        **kwargs,
    ):
        super().__init__()
        self.save_hyperparameters()
        if pretrained == PRESET_MODEL:
            self.model = DetrForObjectDetection.from_pretrained(
                pretrained,
                revision="no_timm",
                num_labels=n_classes,
                ignore_mismatched_sizes=True,
                **kwargs,
            )
        else:
            self.model = DetrForObjectDetection.from_pretrained(
                pretrained,
                **kwargs,
            )

        self.lr = lr
        self.lr_backbone = lr_backbone
        self.weight_decay = weight_decay
        self.processor = DetrImageProcessor.from_pretrained(PRESET_MODEL)

    def forward(self, pixel_values, pixel_mask):
        outputs = self.model(
            pixel_values=pixel_values,
            pixel_mask=pixel_mask,
        )
        return outputs

    def common_step(self, batch, batch_idx):
        pixel_values = batch["pixel_values"].to(self.device)
        pixel_mask = batch["pixel_mask"].to(self.device)
        labels = [{k: v.to(self.device) for k, v in t.items()} for t in batch["labels"]]

        # forward pass
        outputs = self.model(
            pixel_values=pixel_values,
            pixel_mask=pixel_mask,
            labels=labels,
        )

        # calculate loss
        loss = outputs.loss
        loss_dict = outputs.loss_dict

        return loss, loss_dict, outputs, labels

    def training_step(self, batch, batch_idx):
        loss, loss_dict, _, _ = self.common_step(batch, batch_idx)
        self.log("train_loss", loss)
        for k, v in loss_dict.items():
            self.log("train_" + k, v.item())

        return loss

    def validation_step(self, batch, batch_idx):
        loss, loss_dict, _, _ = self.common_step(batch, batch_idx)
        self.log("val_loss", loss)
        for k, v in loss_dict.items():
            self.log("val_" + k, v.item())

        return loss

    def predict_step(self, batch, batch_idx):
        """
        return
        ---
        a list of supervision detection
        """
        loss, loss_dict, outputs, labels = self.common_step(batch, batch_idx)
        results = self.processor.post_process_object_detection(
            outputs, target_sizes=[l["size"] for l in labels]
        )
        preds = [sv.Detections.from_transformers(o) for o in results]
        return preds

    def configure_optimizers(self):
        param_dicts = [
            # params for transformer
            {
                "params": [
                    p
                    for n, p in self.named_parameters()
                    if "backbone" not in n and p.requires_grad
                ]
            },
            # params for resnet
            {
                "params": [
                    p
                    for n, p in self.named_parameters()
                    if "backbone" in n and p.requires_grad
                ],
                "lr": self.lr_backbone,
            },
        ]
        optimizer = torch.optim.AdamW(
            param_dicts,
            lr=self.lr,
            weight_decay=self.weight_decay,
        )

        return optimizer


def convert_to_xywh(boxes):
    xmin, ymin, xmax, ymax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)


def prepare_for_coco_detection(predictions):
    coco_results = []
    for original_id, prediction in predictions.items():
        if len(prediction) == 0:
            continue

        boxes = prediction["boxes"]
        boxes = convert_to_xywh(boxes).tolist()
        scores = prediction["scores"].tolist()
        labels = prediction["labels"].tolist()

        coco_results.extend(
            [
                {
                    "image_id": original_id,
                    "category_id": labels[k],
                    "bbox": box,
                    "score": scores[k],
                }
                for k, box in enumerate(boxes)
            ]
        )
    return coco_results



from transformers import DeiTModel, DeiTConfig
import numpy as np
import torch
import torch.nn as nn
import torchvision
import lightning as l
from pyniche.models.mlp import NicheMLP
# how to freeze the model:
# https://lightning.ai/docs/pytorch/stable/advanced/transfer_learning.html
# how to log hyperparameters:
# https://lightning.ai/docs/pytorch/latest/extensions/logging.html#logging-hyperparameters


class NicheDeiTModule(l.LightningModule):
    def __init__(
        self,
        pretrained: str,  # facebook/deit-base-distilled-patch16-224
        lr: float = 1e-4,
        mlp_hidden_features: int = 256,
        mlp_n_layers: int = 3,
        mlp_out_features: int = 11,
        freeze_ext: bool = True,
        batch: int = -1, # for logging only
    ):
        super().__init__()
        self.save_hyperparameters()
        self.pretrained = pretrained
        self.freeze_ext = freeze_ext
        self.lr = lr
        self.mlp_hidden_features = mlp_hidden_features
        self.mlp_n_layers = mlp_n_layers
        self.mlp_out_features = mlp_out_features
        self.batch = batch
        self.best_val_loss = np.inf
        # model configuration
        self.deit, self.mlp = self.configure_model()
        self.loss_func = nn.CrossEntropyLoss()
        # prepare for training
        self = self.float()
        self.example_input_array = torch.rand(1, 3, 224, 224)
        
    def forward(self, x):
        out = self.deit(x)
        out = self.mlp(out.pooler_output)
        return out

    # behavior before the training starts
    def on_fit_start(self):
        # print all trainable parameters
        print("--- NICHE MODULE MESSAGE ---")
        print("Trainable parameters:")
        for name, param in self.named_parameters():
            if param.requires_grad:
                print(name)
        # # log
        # self.logger.log_hyperparams(
        #     self.hparams, 
        #     {"cross_entropy_loss": 9999}
        # )
    
    def on_fit_end(self):
        """
        log the best val_loss        
        """
        best_val_loss = self.trainer.checkpoint_callback.best_model_score.item()
        # "hp_metric" is the default name for TensorBoardLogger
        self.logger.log_hyperparams(
            self.hparams, 
            {"cross_entropy_loss": best_val_loss}
        )
    def training_step(self, batch, batch_idx):
        x, y = self.extract_batch(batch)
        loss = self.loss_func(self(x), y)
        self.logger.experiment.add_scalar("Loss/Train", loss, self.global_step)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = self.extract_batch(batch)
        loss = self.loss_func(self(x), y)
        self.log("val_loss", loss)
        self.logger.experiment.add_scalar("Loss/Val", loss, self.global_step)
        return loss

    def test_step(self, batch, batch_idx):
        x, y = self.extract_batch(batch)
        loss = self.loss_func(self(x), y)
        return loss

    def predict_step(self, batch, batch_idx):
        x, y = self.extract_batch(batch)
        return self(x)  # return predictions

    def extract_batch(self, batch):
        x, y = batch["x"], batch["y"]
        x = x.float()
        y = y.squeeze(1).long() # from (batch, 1) to (batch)
        return x, y

    def configure_optimizers(self):
        trainable_params = filter(lambda p: p.requires_grad, self.parameters())
        optimizer = torch.optim.Adam(trainable_params, lr=self.lr)
        return optimizer

    def configure_model(self):
        # config
        deit_config = DeiTConfig.from_pretrained(self.pretrained)
        deit_config.output_attentions = True
        # modify DeiT
        deit = DeiTModel(deit_config)
        layers = list(deit.children())
        in_features = layers[-1].dense.in_features
        deit.pooler.dense = nn.Identity()
        deit.pooler.activation = nn.Identity()
        # MLP
        mlp = NicheMLP(
            in_features=in_features,
            out_features=self.mlp_out_features,
            hidden_features=self.mlp_hidden_features,
            n_layers=self.mlp_n_layers,
        )
        # freeze the model
        for param in deit.parameters():
            param.requires_grad = not self.freeze_ext
        for param in mlp.parameters():
            param.requires_grad = True
        # return
        return deit, mlp

    def unfreeze_ext(self):
        for param in self.deit.parameters():
            param.requires_grad = True

# # TEST
# PRETRAINED = "facebook/deit-base-distilled-patch16-224"
# ts = torch.rand(1, 3, 224, 224)

# deito = DeiTModel.from_pretrained(PRETRAINED)
# deito.pooler.dense = nn.Identity()

# outo = deito(ts)
# outo.keys()
# outo.last_hidden_state.shape
# outo.pooler_output.shape



import torch.nn as nn

class NicheMLP(nn.Module):
    def __init__(self, in_features, out_features, hidden_features, n_layers):
        super().__init__()

        layers = []
        layers.append(nn.Linear(in_features, hidden_features))
        layers.append(nn.ReLU())
        for _ in range(n_layers - 1):
            layers.append(nn.Linear(hidden_features, hidden_features))
            layers.append(nn.ReLU())
        layers.append(nn.Linear(hidden_features, out_features))
        
        self.mlp = nn.Sequential(*layers)

    def forward(self, x):
        return self.mlp(x)


import os
import optuna


class Optimizer:
    def __init__(
        self,
        dir_db,
        name_db="optimization.db",
        name_study="optimization",
        n_trials=500,
    ):
        path_db = "sqlite:///%s" % os.path.join(dir_db, name_db)
        if not os.path.exists(dir_db):
            os.makedirs(dir_db)

        self.n_trials = n_trials
        self.study = optuna.create_study(
            storage=path_db,
            study_name=name_study,
            direction="minimize",
            load_if_exists=True,
        )

    def run(
        self,
        objective: callable,
    ):
        self.study.optimize(objective, n_trials=self.n_trials)
        return self.study.best_params, self.study.best_value



import torch


def resize_bbox(bbox, target_size):
    """
    turn a bbox from one size to another

    param
    ---
    bbox: all floating numbers: [x, y, w, h]
    target_size: a tensor of [h, w] ***** very important! *****

    return
    ---
    a new bbox with the same aspect ratio, but resized to fit the target size
    """
    bx, by, bw, bh = bbox
    th, tw = target_size
    new_x = bx * tw
    new_y = by * th
    new_w = bw * tw
    new_h = bh * th
    return torch.tensor([new_x, new_y, new_w, new_h])


def xywh2xyxy(xywh, in_xy="center", img_size=None):
    """
    param
    ---
    xywh: all floating numbers: [x, y, w, h] (range: 0-1)
    in_xy: "center" (yolo) or "top-left" (coco) of the input bbox xy
        - Note that the output from DETR is in "center" format
    img_size: (w, h) of the image
    return
    ---
    xyxy: all floating numbers: [x1, y1, x2, y2] from top-left to bottom-right
    """
    x, y, w, h = xywh
    x = float(x)
    y = float(y)
    w = float(w)
    h = float(h)
    if in_xy == "center":
        dw, dh = w / 2, h / 2
        x1 = x - dw
        y1 = y - dh
        x2 = x + dw
        y2 = y + dh
    elif in_xy == "top-left":
        x1 = x
        y1 = y
        x2 = x + w
        y2 = y + h
    if img_size:
        img_w, img_h = img_size
        x1, x2 = x1 * img_w, x2 * img_w
        y1, y2 = y1 * img_h, y2 * img_h
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)

    return torch.tensor([x1, y1, x2, y2])



import unittest
import niche_cv as nc
from niche_cv.detection.models.detr import Niche_Detr
import os

ROOT = os.path.dirname(os.path.dirname(__file__))
# output
DIR_TEST = os.path.join(ROOT, "test")
DIR_OUT = os.path.join(DIR_TEST, "out")
# module
DIR_MODULE = os.path.join(ROOT, "niche_cv")
DIR_DATA = os.path.join(DIR_MODULE, "detection", "data", "balloon")


class TestDeTR(unittest.TestCase):
    def setUp(self):
        self.model = Niche_Detr(
            path_model="facebook/detr-resnet-50",
            dir_out=DIR_OUT,
            name_task="test",
            device="CUDA",
        )
        print("--- Finish setup ---")
        print(self.model)

    def test_train(self):
        self.model.train(path_data=DIR_DATA, batch=16, epochs=1)



import numpy as np
import supervision as sv
import matplotlib.pyplot as plt


def vis_detections(
    image,
    detections,
    text="object",
    thickness=2,
    figsize=(20, 10),
):
    """
    img: PIL.Image for the annotation to be drawn on
    detections: sv.Detection
    text: text annotations, e.g., "balloon". If None, no text is drawn
    """

    ann_bbox = sv.BoundingBoxAnnotator(thickness=thickness)
    ann_text = sv.LabelAnnotator(text_position=sv.Position.BOTTOM_LEFT)

    # draw predictions
    att_frame_pred = ann_bbox.annotate(np.array(image), detections)
    if text:
        text_ann = [f"{text}" for _ in range(len(detections))]
        att_frame_pred = ann_text.annotate(
            att_frame_pred,
            detections,
            labels=text_ann,
        )
    plt.figure(figsize=figsize)
    plt.imshow(att_frame_pred)



import supervision as sv


def from_yolo(metrics):
    # source
    # ultralytics.utils.metrics.DetMetrics
    """
    args
    ----
        metrics: dict
            metrics from model.val()

    return
    ------
        json: dict
    """
    # metrics
    map5095 = metrics.box.map.round(4)
    map50 = metrics.box.map50.round(4)
    precision = metrics.box.p[0].round(4)
    recall = metrics.box.r[0].round(4)
    f1 = metrics.box.f1[0].round(4)
    # confusion matrix
    conf_mat = metrics.confusion_matrix.matrix  # conf=0.25, iou_thres=0.45
    n_all = conf_mat[:, 0].sum()
    n_fn = conf_mat[1, 0].sum()
    n_fp = conf_mat[0, 1].sum()
    # write json
    json_out = dict(
        map5095=map5095,
        map50=map50,
        precision=precision,
        recall=recall,
        f1=f1,
        n_all=int(n_all),
        n_fn=int(n_fn),  # false negative
        n_fp=int(n_fp),
    )
    return json_out


def from_sv(preds, obs):
    """
    both preds and obs are lists of sv.Detections
    """
    metric_matrix = compute_confusion_matrix(preds, obs)
    SV_MAP = sv.MeanAveragePrecision.from_detections(
        predictions=preds,
        targets=obs,
    )
    # write json
    json_out = dict(
        map5095=SV_MAP.map50_95,
        map50=SV_MAP.map50,
        precision=metric_matrix["precision"],
        recall=metric_matrix["recall"],
        f1=2
        * (metric_matrix["precision"] * metric_matrix["recall"])
        / (metric_matrix["precision"] + metric_matrix["recall"]),
        n_all=len(obs),
        n_fn=metric_matrix["fn"],
        n_fp=metric_matrix["fp"],
    )
    return json_out


def compute_confusion_matrix(
    preds,
    obs,
    conf_threshold=0.25,
    iou_threshold=0.45,
):
    confusion_matrix = sv.ConfusionMatrix.from_detections(
        predictions=preds,
        targets=obs,
        classes=["balloon"],
        conf_threshold=conf_threshold,
        iou_threshold=iou_threshold,
    )
    tp = confusion_matrix.matrix[0][0]
    fn = confusion_matrix.matrix[0][1]
    fp = confusion_matrix.matrix[1][0]
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)
    return dict(
        tp=int(tp),
        fn=int(fn),
        fp=int(fp),
        precision=precision,
        recall=recall,
    )



import argparse


def main():
    parser = argparse.argumentParser(description="My data making program")
    parser.add_argument("-seg", type=str, help="An example argument")

    args = parser.parse_args()

    # Your code here
    print(args.seg)


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pyniche",
    version="0.0.9",
    url="https://github.com/Niche-Squad/pyniche",
    author="James Chen",
    author_email="niche@vt.edu",
    description="An AI Library for Niche Squad",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        # data manipulation
        "pandas",
        "numpy",
        "scipy",
        "scikit-learn",
        "datasets",
        "supervision",
        # modeling
        "optuna",
        "albumentations",
        # deep learning frameworks
        "ultralytics",
        "lightning",
        "transformers",
        "torch",
        "torchvision",
        "torchaudio",
        # visualization
        "seaborn",
        "matplotlib",
    ],
    entry_points={"console_scripts": ["niche=pyniche.show:main"]},
)



import os
import json
import time
from ultralytics import YOLO


class Niche_YOLO:
    def __init__(self, path_model):
        """
        path_model: str
            path to model.pt
            yolov8n.pt
            yolov8s.pt
            yolov8m.pt
            yolov8l.pt
            yolov8x.pt

        Folder structure
        models/
            <path_model>.pt (options: yolov8n.pt yolov8s.pt yolov8m.pt yolov8l.pt yolov8x.pt)
        out/
            train/
                <name_task>/
                    weights/
                        best.pt
                    ...
            val/
                <name_task>/
                    results.json
                    ...

        """
        # attributes
        self.model = YOLO(path_model)
        print("model %s loaded" % path_model)

    def train(
        self,
        path_yaml,  # path to data.yaml
        name_task,
        batch=16,
        epochs=100,
        device="cuda",  # "mps", or cpu
    ):
        """
        <project>/
            <name_task>/
                weights/
                    best.pt
                ...
        """
        start = time.time()
        self.model.train(
            data=path_yaml,
            batch=batch,
            epochs=epochs,
            device=device,
            project=".",
            name=name_task,
            exist_ok=True,
        )
        time_elapsed = time.time() - start
        best_model = os.path.join(name_task, "weights", "best.pt")
        self.model = YOLO(best_model)
        return time_elapsed

    def evaluate(
        self,
        name_task,
        split="test",
        device="cuda",
    ):
        metrics = self.model.val(
            split=split,
            device=device,
            project=".",
            name=name_task,
            exist_ok=True,
        )
        return ext_metrics(metrics)


def ext_metrics(metrics):
    # source
    # ultralytics.utils.metrics.DetMetrics
    """
    args
    ----
        metrics: dict
            metrics from model.val()

    return
    ------
        json: dict
    """
    # metrics
    map5095 = metrics.box.map.round(4)
    map50 = metrics.box.map50.round(4)
    precision = metrics.box.p[0].round(4)
    recall = metrics.box.r[0].round(4)
    f1 = metrics.box.f1[0].round(4)
    # confusion matrix
    conf_mat = metrics.confusion_matrix.matrix  # conf=0.25, iou_thres=0.45
    n_all = conf_mat[:, 0].sum()
    n_fn = conf_mat[1, 0].sum()
    n_fp = conf_mat[0, 1].sum()
    # write json
    json_out = dict(
        map5095=map5095,
        map50=map50,
        precision=precision,
        recall=recall,
        f1=f1,
        n_all=int(n_all),
        n_fn=int(n_fn),  # false negative
        n_fp=int(n_fp),
    )
    return json_out



import optuna
import os

def init_optimizer(
        dir_db: str,
        name_db: str = "optimization.db",
        name_study: str = "optimization",
    ):
        path_db = "sqlite:///%s" % os.path.join(dir_db, name_db)
        if not os.path.exists(dir_db):
            os.makedirs(dir_db)

        try:
            study = optuna.load_study(
                storage=path_db,
                study_name=name_study,
            )
        except KeyError:
            # if study does not exist, create it
            study = optuna.create_study(
                storage=path_db,
                study_name=name_study,
                direction="minimize",
            )
        return study


import optuna
import os


class NicheOptimizer:
    def __init__(
        self,
        dir_db: str,
        name_db: str = "optimization.db",
        name_study: str = "optimization",
    ):
        path_db = "sqlite:///%s" % os.path.join(dir_db, name_db)
        if not os.path.exists(dir_db):
            os.makedirs(dir_db)

        try:
            study = optuna.load_study(
                storage=path_db,
                study_name=name_study,
            )
        except KeyError:
            # if study does not exist, create it
            study = optuna.create_study(
                storage=path_db,
                study_name=name_study,
                direction="minimize",
            )
        self.optimizer = study

    def optimize(
        self,
        objective: callable,
        n_trials: int = 500,
    ):
        self.optimizer.optimize(objective, n_trials=n_trials)


# def objective(trial, dataset):
#     # Suggest hyperparameter values using the trial object
#     learning_rate = trial.suggest_float("learning_rate", 1e-6, 1e-2, log=True)
#     batch_size = trial.suggest_categorical("batch_size", [4, 8, 16, 32, 64])
#     n_blocks = trial.suggest_int("n_blocks", 1, 6)
#     print("The parameters are: ", learning_rate, batch_size, n_blocks)

#     # Train the model with the given hyperparameters and return the validation loss
#     test_loss = sperm_teller(
#         dataset=dataset,
#         epochs=40,
#         lr=learning_rate,
#         batch=batch_size,
#         n_blocks=n_blocks,
#     )
#     return test_loss

# study.optimize(lambda trial: objective(trial, dataset), n_trials=500)
# study.optimize(objective, n_trials=50)



"""
This module is used to process HuggingFace's DETR COCO dataset to be used with LightningDataModule.
"""

# native
import numpy as np

# local imports
from pyniche.data.base import BaseDataModule
from pyniche.data.bbox import resize_bbox, xywh2xyxy

# deep learning
import torch
from transformers import DetrImageProcessor
import albumentations as A
import supervision as sv

# CONSTANTS
PROCESSOR = DetrImageProcessor.from_pretrained("facebook/detr-resnet-50")


class DetectDataModule(BaseDataModule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_labels(self, split):
        """
        extract labels from dataloader and flatten them to (N, )
        """
        dataloader = self.get_dataloader(split)
        labels = []
        for batch in iter(dataloader):
            labels += [{k: v for k, v in t.items()} for t in batch["labels"]]
        return labels

    def get_pixels(self, split):
        """
        extract padded pixels from dataloader and flatten them to (N, )
        """
        dataloader = self.get_dataloader(split)
        pixels = []
        for batch in iter(dataloader):
            pixels += batch["pixel_values"]
        return pixels

    def get_images(self, split):
        """
        extract images from dataloader and flatten them to (N, )
        """
        dataset = self.get_dataset(split)
        return dataset["image"]

    def get_detections(self, split):
        """
        transform the labels to sv.Detections
        """
        labels = self.get_labels(split)
        return [label2detect(l) for l in labels]

    def _collate_fn(self, batch):
        """
        Turn a list of structs into a struct of arrays.

        param
        ---
        batch
            - sample 1
                - pixel_values
                - pixel_mask
                - labels
                    - image_id
                    - annotations
                        - id
                        - image_id
                        - category_id
                        - bbox
                        - iscrowd
                        - area
                        - segmentation
            - sample 2
                - pixel_values
                - pixel_mask
                - labels

        return
        ---
        a dict
            - pixel_values
                - sample 1
                - sample 2
                ...
            - pixel_mask
                - sample 1
                - sample 2
                ...
            - labels
                - sample 1
                - sample 2
                ...
        """
        new_batch = {}
        new_batch["pixel_values"] = torch.stack(
            [item["pixel_values"] for item in batch]
        )
        new_batch["pixel_mask"] = torch.stack([item["pixel_mask"] for item in batch])
        new_batch["labels"] = [item["labels"] for item in batch]
        return new_batch

    def _collate_fn_train(self, batch):
        return self._collate_fn(batch)

    def _collate_fn_val(self, batch):
        return self._collate_fn(batch)

    def _collate_fn_test(self, batch):
        return self._collate_fn(batch)

    def _transform_train(self, examples):
        """
        input
        ---
        examples (a struct of arrays, SOA)
            image
                image 1
                image 2
                ...
            image_id
                image_id 1
                image_id 2
                ...
            ...
            annotations
                annotation 1
                    id
                        id 1
                        id 2
                        ...
                    image_id
                        image_id 1
                        image_id 1
                        ...
                    ...
                annotation 2
                    ...
                ...

        output
        ---
        batch (an array of structs, AOS)
            example 1
                pixel_values
                pixel_mask
                labels
                    size
                    image_id
                    class_labels
                        label 1
                        label 2
                        ...
                    boxes
                        box 1
                        box 2
                        ...
                    area
                        area 1
                        area 2
                        ...
                    iscrowd
                        iscrowd 1
                        iscrowd 2
                        ...
                    orig_size
            example 2
            ...
        """
        transform = A.Compose(
            [
                # A.HorizontalFlip(p=0.5),
                # A.VerticalFlip(p=0.5),
                # A.RandomBrightnessContrast(p=1),
            ],
            bbox_params=A.BboxParams(format="coco", label_fields=["category"]),
        )
        batch = process(examples, transform)
        return batch

    def _transform_val(self, examples):
        transform = A.Compose(
            [],
            bbox_params=A.BboxParams(format="coco", label_fields=["category"]),
        )
        batch = process(examples, transform)
        return batch

    def _transform_test(self, examples):
        transform = A.Compose(
            [],
            bbox_params=A.BboxParams(format="coco", label_fields=["category"]),
        )
        batch = process(examples, transform)
        return batch


def process(examples, transform):
    transformed_images = []
    transformed_anns = []
    for image, image_id, annotations in zip(
        examples["image"], examples["image_id"], examples["annotations"]
    ):
        # image
        image = np.array(image.convert("RGB"))[:, :, ::-1]
        # annotations
        out = transform(
            image=image,
            bboxes=annotations["bbox"],
            category=annotations["category_id"],
        )
        ann = {
            "image_id": image_id,
            "annotations": [
                {
                    "id": annotations["id"][i],
                    "image_id": annotations["image_id"][i],
                    "category_id": out["category"][i],
                    "bbox": out["bboxes"][i],
                    "iscrowd": annotations["iscrowd"][i],
                    "area": annotations["area"][i],
                    "segmentation": annotations["segmentation"][i],
                }
                for i in range(len(annotations["id"]))
            ],
        }
        # append
        transformed_images.append(out["image"])
        transformed_anns.append(ann)
    # process the transformed data
    batch = PROCESSOR(
        images=transformed_images,
        annotations=transformed_anns,
        return_tensors="pt",
    )
    return batch


def label2detect(label):
    size = label["size"]
    class_id = label["class_labels"].numpy()
    bbox = [resize_bbox(b, size) for b in label["boxes"]]
    bbox = np.array([xywh2xyxy(b) for b in bbox])
    return sv.Detections(xyxy=bbox, class_id=class_id)



import matplotlib.pyplot as plt
import matplotlib.patches as patches
import random
import numpy as np
import colorsys
import torch
from torchvision.transforms import ToPILImage

import supervision as sv


def tensor_to_pil(img_tensor):
    """
    img_tensor: (C, H, W) tensor
    """
    # turn (C, H, W) to (H, W, C)
    img_tensor = img_tensor.permute(1, 2, 0).numpy()
    # normalize
    img_tensor -= img_tensor.min()
    img_tensor = (img_tensor * 255 / img_tensor.max()).astype(np.uint8)
    img_tensor = img_tensor[..., ::-1]
    return ToPILImage()(img_tensor)


def vis_detect(
    image,
    preds,
    obs=None,
    text="object",
    thickness=2,
    figsize=(20, 10),
):
    """
    img: PIL.Image for the annotation to be drawn on
    preds: sv.Detection
    obs: sv.Detection
    text: text annotations, e.g., "balloon"
    """
    if isinstance(image, torch.Tensor):
        image = tensor_to_pil(image)

    ann_bbox = sv.BoundingBoxAnnotator(thickness=thickness)
    ann_text = sv.LabelAnnotator(text_position=sv.Position.BOTTOM_LEFT)

    # draw predictions
    text_pred = [f"{text}: {conf:0.2f}" for conf in preds.confidence]
    att_frame_pred = ann_bbox.annotate(np.array(image), preds)
    att_frame_pred = ann_text.annotate(att_frame_pred, preds, labels=text_pred)

    if obs is not None:
        text_obs = [f"{text}" for _ in obs]
        att_frame_obs = ann_bbox.annotate(np.array(image), obs)
        att_frame_obs = ann_text.annotate(att_frame_obs, obs, labels=text_obs)
        fig, ax = plt.subplots(1, 2, figsize=figsize)
        ax[0].imshow(att_frame_pred)
        ax[0].set_title("Predictions")
        ax[1].imshow(att_frame_obs)
        ax[1].set_title("Observations")
    else:
        plt.figure(figsize=figsize)
        plt.imshow(att_frame_pred)


# def show_dataset(dataset, idx=None):
#     # COCO API
#     coco = dataset.coco

#     # get image id from idx
#     img_ids = list(coco.imgs.keys())
#     if idx is None:
#         idx = random.randint(0, len(img_ids))
#     img_id = img_ids[idx]
#     image = dataset._load_image(img_id)

#     # Get all distinct categories and create a color for each of them
#     categories = coco.loadCats(coco.getCatIds())
#     num_classes = len(categories)
#     hsv_tuples = [(x / num_classes, 1.0, 1.0) for x in range(num_classes)]
#     colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
#     colors = [(int(r * 255), int(g * 255), int(b * 255)) for r, g, b in colors]
#     random.shuffle(colors)

#     # Display the image
#     plt.figure(figsize=(10, 10))
#     plt.imshow(image)

#     # Add the bounding boxes
#     ax = plt.gca()
#     annotation_ids = coco.getAnnIds(imgIds=img_id, iscrowd=None)
#     annotations = coco.loadAnns(annotation_ids)
#     annotation = annotations[0]
#     annotation["bbox"]
#     annotation["segmentation"][0]

#     for annotation in annotations:
#         # Draw bounding box
#         bbox = annotation["bbox"]
#         box = patches.Rectangle(
#             (bbox[0], bbox[1]),
#             bbox[2],
#             bbox[3],
#             linewidth=2,
#             edgecolor="r",
#             facecolor="none",
#         )
#         ax.add_patch(box)

#         # Draw segmentation
#         if "segmentation" in annotation:
#             category_id = annotation["category_id"]
#             color = (
#                 colors[categories.index(category_id)]
#                 if category_id in categories
#                 else "r"
#             )
#             for seg in annotation["segmentation"]:
#                 poly = np.array(seg).reshape((int(len(seg) / 2), 2))
#                 polygon = patches.Polygon(
#                     poly, linewidth=1, edgecolor=color, facecolor=color, alpha=0.5
#                 )
#                 ax.add_patch(polygon)

#         # Draw label
#         label = coco.loadCats(annotation["category_id"])[0]["name"]
#         plt.text(
#             bbox[0],
#             bbox[1] - 10,
#             label,
#             color="white",
#             fontsize=12,
#             bbox={"facecolor": "red", "alpha": 0.5},
#         )
